#
# Chatstream -- Streaming Web Chat Server Without Java
#
# Copyright (C) 1997 by Karl Lehenbauer, All Rights Reserved.
#
# $Id: httpd.tcl,v 1.11 1997/03/27 02:46:22 karl Exp $
#

source config

#
# reply_ok -- issue a standard HTTP reply header
#
proc reply_ok {socket {contentType text/html}} {
    puts $socket "HTTP/1.0 200 OK"
    puts $socket "Server: NeoWebScript-SA/3.1"
    puts $socket "Content-type: $contentType\n"
}

#
# url_connectinfo -- reply a formatted webpage containing connection info
#
proc url_connectinfo {arrayName} {
    upvar #0 $arrayName connection
    set socket $connection(socket)

    reply_ok $socket

    puts $socket "<title>Connection Info</title>"
    puts $socket "<body bgcolor=black text=white>"
    puts $socket "<table border=2>"
    puts $socket "<caption>Connection Info</caption>"
    puts $socket "<tr><th>Variable</th><th>Value</th></tr>"

    foreach var [lsort [array names connection]] {
	set value $connection($var)
	if {$value == ""} {set value "<i>none</i>"}
	puts $socket "<tr><td>$var</td><td>$value</td></tr>"
    }
    puts $socket "</table>"
    close $socket
    unset connection
}

#
# url_pushit -- virtual webpage that begins a push to a frame of stuff
# people type to the chat server
#
proc url_pushit {arrayName} {
    upvar #0 $arrayName connection
    set socket $connection(socket)

    echo "PUSHY PUSHY"
    reply_ok $socket

    global pushSockets
parray pushSockets
    foreach connectName [array names pushSockets] {
echo "linking target to $connectName"
	upvar #0 $connectName target
parray target

	puts $target(socket) "$connection(postHeader)<br>"
	flush $target(socket)
    }
    set ifp [open htdocs/input_window.html]
    unsupported0 $ifp $socket
    close $ifp
    close $socket
    unset connection

}

proc url_pull {arrayName} {
    global bgColor textColor linkColor vlinkColor alinkColor

    upvar #0 $arrayName connection
    global pushSockets

    set socket $connection(socket)

    if {![info exists connection(headerEmitted)]} {
	set connection(headerEmitted) 1

        reply_ok $socket
    }
    set pushSockets($arrayName) ""

    puts $socket "<body bgcolor=$bgColor text=$textColor link=$linkColor vlink=$vlinkColor alink=$alinkColor>"
}

#
# dispatch - handler that's called once we've receive a complete
# 	     request from a web browser.
#
proc dispatch {arrayName} {
    upvar #0 $arrayName connection

    set socket $connection(socket)

    set fileName htdocs$connection(url)
    set urlName url_[crange $connection(url) 1 end]

    if {[info procs $urlName] != ""} {

	$urlName $arrayName
	return

    } elseif {[file readable $fileName] && [file type $fileName] == "file"} {
	reply_ok $socket

	set ifp [open $fileName]
	unsupported0 $ifp $socket
	close $ifp
    } else {
	puts $socket "HTTP/1.0 404 NotFound"
	puts $socket "Server: NeoWebScript-SA/3.1"
	puts $socket "Content-type: text/html\n"
	puts $socket "<title>404 Not Found</title>"
	puts $socket "<h1>404 Not Found</h1>"
	puts $socket "The document you requested <b>$connection(url)</b> was not found."
    }
    close $socket
    unset connection
}

#
# data_available - called when data comes in from a socket connection that 
# is already open
#
proc data_available {arrayName} {
    upvar #0 $arrayName connection
    #echo "data_available on $arrayName"

    set socket $connection(socket)

    if {[gets $socket line] < 0} {
	if {[eof $socket]} {
	    echo "EOF received from socket $socket, BYEBYE"
	    close $socket
	    unset connection
	    return
	}
	echo "called with data but none available on socket $socket"
	return
    }

    if {$line == ""} {
	set connection(postHeader) ""
	after idle "dispatch $arrayName"
	return
    }

    if {$connection(command) == ""} {
	echo [regexp {([^ ]*) ([^ ]*) (.*)} $line dummy \
	    connection(command) connection(url) connection(httpType)]
	return
    }

    if {[info exists connection(postHeader)]} {
	lappend connection(postHeader) $line
	return
    }

    if {![regexp {([^:]*): (.*)} $line dummy key value]} {
	echo "unparseable line: $line"
	return
    }
    set connection(http_$key) $value
}

#
# handle_connection_request - called when a connection request is received
#                             on our server socket
#
proc handle_connection_request {socket ipAddress port} {
    echo "connect request from $ipAddress, port $port, as socket $socket"

    upvar #0 connect_${socket}_${port} connection
    set connection(socket) $socket
    set connection(command) ""

    fconfigure $socket -blocking 0 -translation {auto crlf}
    fileevent $socket readable "data_available connect_${socket}_${port}"
}

#
# setup_server - sets up the server socket to do stuff
#
proc setup_server {} {
    global port
    socket -server handle_connection_request $port
}

#
# radar - do something every so often
#
proc radar {} {
     echo "RADAR"
     foreach connectionName [info globals connect_*] {
	 echo "radaring $connectionName"
	 upvar #0 $connectionName connection
	 puts $connection(socket) "beep<br>"
	 flush $connection(socket)
     }
     after 60000 radar
}

#
# bgerror - sort of handle errors from code called by events
#
proc bgerror {args} {
    global errorInfo errorCode

    echo "Error: $args"
    echo "Error Code: $errorCode"
    echo "Error Info: $errorInfo"
}

proc announce {} {
    global port

    echo "NeoWebScript Chatstream Streaming Chat Server is On-Line at http://[id host]:$port/"
}

proc doit {} {
    setup_server

    announce

    after 5000 radar

    vwait die
}

if !$tcl_interactive doit
